#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


# TODO: share codes
def generate_pie_data(df, x: str, y: str, topk: Optional[int] = None):
    ret = df.copy()
    ret.set_index(x, inplace=True)
    ret.sort_values(y, inplace=True, ascending=False)
    if topk is not None:
        a = ret.iloc[:topk]
        b = ret.iloc[topk:].agg(others=(y, "sum"))
        ret = pd.concat([a, b])
    ret = ret[y]

    return ret


def compare_with_sess(
    sess: Session,
    other_sess: Session,
    force: bool = False,
):
    logger.info("Comparing ISEs between sessions...")
    artifacts = sess.artifacts
    artifacts_ = other_sess.artifacts
    pc2bb_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb")
    assert len(pc2bb_artifacts) == 1
    pc2bb_artifact = pc2bb_artifacts[0]
    pc2bb_df = pc2bb_artifact.df
    pc2bb_artifacts_ = filter_artifacts(artifacts_, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb")
    assert len(pc2bb_artifacts_) == 1
    pc2bb_artifact_ = pc2bb_artifacts_[0]
    pc2bb_df_ = pc2bb_artifact_.df

    # symbol_map_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_map"
    # )
    # symbol_map_df = None
    # if len(symbol_map_artifacts) > 0:
    #     assert len(symbol_map_artifacts) == 1
    #     symbol_map_artifact = symbol_map_artifacts[0]
    #     symbol_map_df = symbol_map_artifact.df

    # instrs_hist_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "instrs_hist"
    # )
    # instrs_hist_df = None
    # if len(instrs_hist_artifacts) > 0:
    #     assert len(instrs_hist_artifacts) == 1
    #     instrs_hist_artifact = instrs_hist_artifacts[0]
    #     instrs_hist_df = instrs_hist_artifact.df

    # opcodes_hist_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "opcodes_hist"
    # )
    # opcodes_hist_df = None
    # if len(opcodes_hist_artifacts) > 0:
    #     assert len(opcodes_hist_artifacts) == 1
    #     opcodes_hist_artifact = opcodes_hist_artifacts[0]
    #     opcodes_hist_df = opcodes_hist_artifact.df

    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new"
    )
    llvm_bbs_df = None
    if len(llvm_bbs_artifacts) > 0:
        assert len(llvm_bbs_artifacts) == 1
        llvm_bbs_artifact = llvm_bbs_artifacts[0]
        llvm_bbs_df = llvm_bbs_artifact.df.copy()
    llvm_bbs_artifacts_ = filter_artifacts(
        artifacts_, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new"
    )
    llvm_bbs_df_ = None
    if len(llvm_bbs_artifacts_) > 0:
        assert len(llvm_bbs_artifacts_) == 1
        llvm_bbs_artifact_ = llvm_bbs_artifacts_[0]
        llvm_bbs_df_ = llvm_bbs_artifact_.df.copy()

    plots_dir = sess.directory / "plots"
    plots_dir.mkdir(exist_ok=True)
    # TODO: use threshold

    if pc2bb_df is not None and pc2bb_df is not None:
        # TODO: share codes with pie
        runtime_df = pc2bb_df.copy()
        runtime_df_ = pc2bb_df_.copy()
        runtime = runtime_df["weight"].sum()
        runtime_ = runtime_df_["weight"].sum()
        runtime_rel = runtime / runtime_
        # print("runtime_rel", runtime_rel)
        # input("!")

        def helper(x):
            if x is None:
                return "?"
            if isinstance(x, set):
                # assert len(x) == 1
                assert len(x) > 0
                # Only pick first element if alias exists
                return list(x)[0]
            return x

        runtime_df["func_name"] = runtime_df["func_name"].apply(helper)
        runtime_df_["func_name"] = runtime_df_["func_name"].apply(helper)

        if runtime_rel <= 1:
            temp = pd.DataFrame([{"func_name": "DIFF", "weight": (1 - runtime_rel) * runtime_}])
            runtime_df = pd.concat([runtime_df, temp])
            runtime_df["rel_weight"] = runtime_df["weight"] / runtime_
            temp_ = pd.DataFrame([{"func_name": "DIFF", "weight": 0, "rel_weight": 0}])
            runtime_df_ = pd.concat([runtime_df_, temp_])
        else:
            # TODO: check if ok?
            temp_ = pd.DataFrame([{"func_name": "DIFF", "weight": (1 - (1 / runtime_rel)) * runtime}])
            runtime_df_ = pd.concat([runtime_df_, temp_])
            runtime_df_["rel_weight"] = runtime_df_["weight"] / runtime
            temp = pd.DataFrame([{"func_name": "DIFF", "weight": 0, "rel_weight": 0}])
            runtime_df = pd.concat([runtime_df, temp])

        # print("runtime_df", runtime_df)
        # print("runtime_df_", runtime_df_)

        runtime_df = runtime_df[["func_name", "rel_weight"]]
        runtime_df = runtime_df.groupby("func_name", as_index=False, dropna=False).sum()
        runtime_per_func_data = generate_pie_data(runtime_df, x="func_name", y="rel_weight")
        runtime_df_ = runtime_df_[["func_name", "rel_weight"]]
        runtime_df_ = runtime_df_.groupby("func_name", as_index=False, dropna=False).sum()
        runtime_per_func_data_ = generate_pie_data(runtime_df_, x="func_name", y="rel_weight")
        # print("runtime_per_func_data", runtime_per_func_data)
        # print("runtime_per_func_data_", runtime_per_func_data_)
        merged_runtime_per_func_data = runtime_per_func_data.to_frame(name="rel_weight").merge(
            runtime_per_func_data_.to_frame(name="rel_weight"),
            on="func_name",
            how="outer",
            suffixes=("", "_"),
            # indicator=True,
        )
        # print("merged_runtime_per_func_data", merged_runtime_per_func_data)
        merged_runtime_per_func_data.sort_values("rel_weight", inplace=True, ascending=False)
        merged_runtime_per_func_data["diff"] = (
            (merged_runtime_per_func_data["rel_weight"] - merged_runtime_per_func_data["rel_weight_"])
            .fillna(0)
            .round(5)
        )
        merged_runtime_per_func_data["diff_rel"] = (
            (merged_runtime_per_func_data["diff"] / merged_runtime_per_func_data["rel_weight_"]).fillna(0).round(5)
        )
        merged_runtime_per_func_data = merged_runtime_per_func_data[merged_runtime_per_func_data["diff"] != 0]
        # print("merged_runtime_per_func_data", merged_runtime_per_func_data)
        # input(">>>")
        attrs = {}
        artifact = TableArtifact("compare_runtime_per_func", merged_runtime_per_func_data, attrs=attrs)
        sess.add_artifact(artifact, override=force)
        if llvm_bbs_df is not None and llvm_bbs_df_ is not None:
            llvm_bbs_df["func_bb"] = llvm_bbs_df["func_name"] + "-" + llvm_bbs_df["bb_name"]
            llvm_bbs_df_["func_bb"] = llvm_bbs_df_["func_name"] + "-" + llvm_bbs_df_["bb_name"]
            if runtime_rel <= 1:
                temp = pd.DataFrame([{"func_bb": "DIFF", "weight": (1 - runtime_rel) * runtime_}])
                llvm_bbs_df = pd.concat([llvm_bbs_df, temp])
                llvm_bbs_df["rel_weight"] = llvm_bbs_df["weight"] / runtime_
                temp_ = pd.DataFrame([{"func_bb": "DIFF", "weight": 0, "rel_weight": 0}])
                llvm_bbs_df_ = pd.concat([llvm_bbs_df_, temp_])
            else:
                # TODO: check if ok?
                temp_ = pd.DataFrame([{"func_bb": "DIFF", "weight": (1 - (1 / runtime_rel)) * runtime}])
                llvm_bbs_df_ = pd.concat([llvm_bbs_df_, temp_])
                llvm_bbs_df_["rel_weight"] = llvm_bbs_df_["weight"] / runtime
                temp = pd.DataFrame([{"func_bb": "DIFF", "weight": 0, "rel_weight": 0}])
                llvm_bbs_df = pd.concat([llvm_bbs_df, temp])
            runtime_per_llvm_bb_data = generate_pie_data(
                llvm_bbs_df,
                x="func_bb",
                y="rel_weight",
            )
            runtime_per_llvm_bb_data_ = generate_pie_data(
                llvm_bbs_df_,
                x="func_bb",
                y="rel_weight",
            )
            # print("runtime_per_llvm_bb_data", runtime_per_llvm_bb_data)
            # print("runtime_per_llvm_bb_data_", runtime_per_llvm_bb_data_)
            merged_runtime_per_llvm_bb_data = runtime_per_llvm_bb_data.to_frame(name="rel_weight").merge(
                runtime_per_llvm_bb_data_.to_frame(name="rel_weight"),
                on="func_bb",
                how="outer",
                suffixes=("", "_"),
                # indicator=True,
            )
            merged_runtime_per_llvm_bb_data.sort_values("rel_weight", inplace=True, ascending=False)
            merged_runtime_per_llvm_bb_data["diff"] = (
                (merged_runtime_per_llvm_bb_data["rel_weight"] - merged_runtime_per_llvm_bb_data["rel_weight_"])
                .fillna(0)
                .round(5)
            )
            merged_runtime_per_llvm_bb_data["diff_rel"] = (
                (merged_runtime_per_llvm_bb_data["diff"] / merged_runtime_per_llvm_bb_data["rel_weight_"])
                .fillna(0)
                .round(5)
            )
            merged_runtime_per_llvm_bb_data = merged_runtime_per_llvm_bb_data[
                merged_runtime_per_llvm_bb_data["diff"] != 0
            ]
            # print("merged_runtime_per_llvm_bb_data", merged_runtime_per_llvm_bb_data)
            # input(">>>")
            attrs = {}
            artifact = TableArtifact(
                "compare_runtime_per_llvm_bb",
                merged_runtime_per_llvm_bb_data,
                attrs=attrs,
            )
            sess.add_artifact(artifact, override=force)

    mem_footprint_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_footprint"
    )
    assert len(mem_footprint_artifacts) == 1
    mem_footprint_artifact = mem_footprint_artifacts[0]
    mem_footprint_df = mem_footprint_artifact.df
    mem_footprint_artifacts_ = filter_artifacts(
        artifacts_, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_footprint"
    )
    assert len(mem_footprint_artifacts_) == 1
    mem_footprint_artifact_ = mem_footprint_artifacts_[0]
    mem_footprint_df_ = mem_footprint_artifact_.df

    effective_mem_footprint_artifacts = filter_artifacts(
        artifacts,
        lambda x: x.flags & ArtifactFlag.TABLE and x.name == "effective_mem_footprint",
    )
    eff_mem_footprint_df = None
    if len(effective_mem_footprint_artifacts) > 0:
        assert len(effective_mem_footprint_artifacts) == 1
        effective_mem_footprint_artifact = effective_mem_footprint_artifacts[0]
        eff_mem_footprint_df = effective_mem_footprint_artifact.df
    effective_mem_footprint_artifacts_ = filter_artifacts(
        artifacts_,
        lambda x: x.flags & ArtifactFlag.TABLE and x.name == "effective_mem_footprint",
    )
    eff_mem_footprint_df_ = None
    if len(effective_mem_footprint_artifacts_) > 0:
        assert len(effective_mem_footprint_artifacts_) == 1
        effective_mem_footprint_artifact_ = effective_mem_footprint_artifacts_[0]
        eff_mem_footprint_df_ = effective_mem_footprint_artifact_.df
    if mem_footprint_df is not None and mem_footprint_df_ is not None:
        mem_footprint = mem_footprint_df["bytes"].sum()
        mem_footprint_ = mem_footprint_df_["bytes"].sum()
        mem_footprint_rel = mem_footprint / mem_footprint_
        # print("mem_footprint", mem_footprint)
        # print("mem_footprint_", mem_footprint_)
        # print("mem_footprint_rel", mem_footprint_rel)
        # input("?")
        # print("mem_footprint_df", mem_footprint_df)
        # print("mem_footprint_df_", mem_footprint_df_)
        if mem_footprint_rel <= 1:
            temp = pd.DataFrame([{"func": "DIFF", "bytes": (1 - mem_footprint_rel) * mem_footprint_}])
            mem_footprint_df = pd.concat([mem_footprint_df, temp])
            mem_footprint_df["rel_bytes"] = mem_footprint_df["bytes"] / mem_footprint_
            temp_ = pd.DataFrame([{"func": "DIFF", "bytes": 0, "rel_bytes": 0}])
            mem_footprint_df_ = pd.concat([mem_footprint_df_, temp_])
        else:
            # TODO: check if ok?
            temp_ = pd.DataFrame(
                [
                    {
                        "func": "DIFF",
                        "bytes": (1 - (1 / mem_footprint_rel)) * mem_footprint,
                    }
                ]
            )
            mem_footprint_df_ = pd.concat([mem_footprint_df_, temp_])
            mem_footprint_df_["rel_bytes"] = mem_footprint_df_["bytes"] / mem_footprint
            temp = pd.DataFrame([{"func": "DIFF", "bytes": 0, "rel_bytes": 0}])
            mem_footprint_df = pd.concat([mem_footprint_df, temp])
        # print("mem_footprint_df2", mem_footprint_df)
        # print("mem_footprint_df_2", mem_footprint_df_)
        mem_footprint_per_func_data = generate_pie_data(
            mem_footprint_df,
            x="func",
            y="rel_bytes",
        )
        mem_footprint_per_func_data_ = generate_pie_data(
            mem_footprint_df_,
            x="func",
            y="rel_bytes",
        )
        merged_mem_footprint_per_func_data = mem_footprint_per_func_data.to_frame(name="rel_bytes").merge(
            mem_footprint_per_func_data_.to_frame(name="rel_bytes"),
            on="func",
            how="outer",
            suffixes=("", "_"),
            # indicator=True,
        )
        # print("merged_mem_footprint_per_func_data", merged_mem_footprint_per_func_data)
        merged_mem_footprint_per_func_data.sort_values("rel_bytes", inplace=True, ascending=False)
        merged_mem_footprint_per_func_data["diff"] = merged_mem_footprint_per_func_data[
            "rel_bytes"
        ] - merged_mem_footprint_per_func_data["rel_bytes_"].fillna(0).round(3)
        merged_mem_footprint_per_func_data["diff_rel"] = merged_mem_footprint_per_func_data[
            "diff"
        ] / merged_mem_footprint_per_func_data["rel_bytes_"].fillna(0).round(3)
        merged_mem_footprint_per_func_data = merged_mem_footprint_per_func_data[
            merged_mem_footprint_per_func_data["diff"] != 0
        ]
        # print("merged_mem_footprint_per_func_data", merged_mem_footprint_per_func_data)
        # input(">>>")
        attrs = {}
        artifact = TableArtifact(
            "compare_mem_footprint_per_func",
            merged_mem_footprint_per_func_data,
            attrs=attrs,
        )
        sess.add_artifact(artifact, override=force)
    if eff_mem_footprint_df is not None and eff_mem_footprint_df_ is not None:
        eff_mem_footprint = eff_mem_footprint_df["bytes"].sum()
        eff_mem_footprint_ = eff_mem_footprint_df_["bytes"].sum()
        eff_mem_footprint_rel = eff_mem_footprint / eff_mem_footprint_
        # print("eff_mem_footprint", eff_mem_footprint)
        # print("eff_mem_footprint_", eff_mem_footprint_)
        # print("eff_mem_footprint_rel", eff_mem_footprint_rel)
        # input("?")
        # print("eff_mem_footprint_df", eff_mem_footprint_df)
        # print("eff_mem_footprint_df_", eff_mem_footprint_df_)
        if eff_mem_footprint_rel <= 1:
            temp = pd.DataFrame(
                [
                    {
                        "func": "DIFF",
                        "bytes": (1 - eff_mem_footprint_rel) * eff_mem_footprint_,
                    }
                ]
            )
            eff_mem_footprint_df = pd.concat([eff_mem_footprint_df, temp])
            eff_mem_footprint_df["eff_rel_bytes"] = eff_mem_footprint_df["bytes"] / eff_mem_footprint_
            temp_ = pd.DataFrame([{"func": "DIFF", "bytes": 0, "eff_rel_bytes": 0}])
            eff_mem_footprint_df_ = pd.concat([eff_mem_footprint_df_, temp_])
        else:
            # TODO: check if ok?
            temp_ = pd.DataFrame(
                [
                    {
                        "func": "DIFF",
                        "bytes": (1 - (1 / eff_mem_footprint_rel)) * eff_mem_footprint,
                    }
                ]
            )
            eff_mem_footprint_df_ = pd.concat([eff_mem_footprint_df_, temp_])
            eff_mem_footprint_df_["eff_rel_bytes"] = eff_mem_footprint_df_["bytes"] / eff_mem_footprint
            temp = pd.DataFrame([{"func": "DIFF", "bytes": 0, "eff_rel_bytes": 0}])
            eff_mem_footprint_df = pd.concat([eff_mem_footprint_df, temp])
        # print("eff_mem_footprint_df2", eff_mem_footprint_df)
        # print("eff_mem_footprint_df_2", eff_mem_footprint_df_)

        eff_mem_footprint_per_func_data = generate_pie_data(
            eff_mem_footprint_df,
            x="func",
            y="eff_rel_bytes",
        )
        eff_mem_footprint_per_func_data_ = generate_pie_data(
            eff_mem_footprint_df_,
            x="func",
            y="eff_rel_bytes",
        )
        merged_eff_mem_footprint_per_func_data = eff_mem_footprint_per_func_data.to_frame(name="eff_rel_bytes").merge(
            eff_mem_footprint_per_func_data_.to_frame(name="eff_rel_bytes"),
            on="func",
            how="outer",
            suffixes=("", "_"),
            # indicator=True,
        )
        # print("merged_eff_mem_footprint_per_func_data", merged_eff_mem_footprint_per_func_data)
        merged_eff_mem_footprint_per_func_data.sort_values("eff_rel_bytes", inplace=True, ascending=False)
        merged_eff_mem_footprint_per_func_data["diff"] = merged_eff_mem_footprint_per_func_data[
            "eff_rel_bytes"
        ] - merged_eff_mem_footprint_per_func_data["eff_rel_bytes_"].fillna(0).round(3)
        merged_eff_mem_footprint_per_func_data["diff_rel"] = merged_eff_mem_footprint_per_func_data[
            "diff"
        ] / merged_eff_mem_footprint_per_func_data["eff_rel_bytes_"].fillna(0).round(3)
        merged_eff_mem_footprint_per_func_data = merged_eff_mem_footprint_per_func_data[
            merged_eff_mem_footprint_per_func_data["diff"] != 0
        ]
        # print("merged_eff_mem_footprint_per_func_data", merged_eff_mem_footprint_per_func_data)
        # input(">>>")
        attrs = {}
        artifact = TableArtifact(
            "compare_eff_mem_footprint_per_func",
            merged_eff_mem_footprint_per_func_data,
            attrs=attrs,
        )
        sess.add_artifact(artifact, override=force)

    # attrs = {}

    # artifact = TableArtifact("compare_sess", compare_sess_df, attrs=attrs)
    # sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    assert args.with_session is not None
    session_dir_ = Path(args.with_session)
    assert session_dir_.is_dir(), f"Session dir does not exist: {session_dir}"
    sess_ = Session.from_dir(session_dir_)
    compare_with_sess(
        sess,
        other_sess=sess_,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--with-session", required=True)
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def file_from_symbol_map(func_name: str, symbol_map_df: pd.DataFrame):
    # print("func_name", func_name)
    # print("symbol_map_df", symbol_map_df, symbol_map_df.columns)
    matches = symbol_map_df[symbol_map_df["symbol"] == func_name]
    # print("matches", matches)
    if len(matches) == 0:
        return None
    obj = matches["object_full"].iloc[0]
    # TODO: resolve relative paths using DWARF?
    if obj.endswith(".o"):
        obj = obj[:-2]
    elif obj.endswith(".obj"):
        obj = obj[:-4]
    # print("obj", obj)
    # input("!")
    return obj


def lookup_files(file2funcs_df, func_name):
    matches = set()
    for _, row in file2funcs_df.iterrows():
        func_names = row["func_names"]
        if func_name in func_names:
            file = Path(row["file"])
            matches.add(file.resolve())
    return list(matches)


def choose_bbs(
    sess: Session,
    threshold: float = 0.9,
    min_weight: float = 0.01,
    min_supported_weight: Optional[float] = None,
    min_instrs: Optional[int] = 2,
    max_num: Optional[int] = None,
    force: bool = False,
):
    logger.info("Choosing BBs...")
    artifacts = sess.artifacts
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]
    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new"
    )
    assert len(llvm_bbs_artifacts) == 1
    llvm_bbs_artifact = llvm_bbs_artifacts[0]
    llvm_bbs_df = llvm_bbs_artifact.df.copy()
    ise_potential_per_llvm_bb_artifacts = filter_artifacts(
        artifacts,
        lambda x: x.flags & ArtifactFlag.TABLE and x.name == "ise_potential_per_llvm_bb",
    )
    ise_potential_per_llvm_bb_df = None
    if len(ise_potential_per_llvm_bb_artifacts) > 0:
        assert len(ise_potential_per_llvm_bb_artifacts) == 1
        ise_potential_per_llvm_bb_artifact = ise_potential_per_llvm_bb_artifacts[0]
        ise_potential_per_llvm_bb_df = ise_potential_per_llvm_bb_artifact.df.copy()

    file2funcs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "file2funcs"
    )
    if len(file2funcs_artifacts) > 0:
        assert len(file2funcs_artifacts) == 1
        file2funcs_artifact = file2funcs_artifacts[0]
        file2funcs_df = file2funcs_artifact.df.copy()
        # print("llvm_bbs_df", llvm_bbs_df)
    else:
        file2funcs_df = None

    symbol_map_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_map"
    )
    symbol_map_df = None
    if len(symbol_map_artifacts) > 0:
        assert len(symbol_map_artifacts) == 1
        symbol_map_artifact = symbol_map_artifacts[0]
        symbol_map_df = symbol_map_artifact.df
    sum_weights = 0.0
    choices = []
    for index, row in llvm_bbs_df.sort_values("rel_weight", ascending=False).iterrows():
        func_name = row["func_name"]
        bb_name = row["bb_name"]
        rel_weight = row["rel_weight"]
        num_instrs = row["num_instrs"]
        if pd.isna(rel_weight):
            continue
        if rel_weight < min_weight:
            continue
        if min_supported_weight is not None:
            assert (
                ise_potential_per_llvm_bb_df is not None
            ), "Run isaac_toolkit.generate.ise.check_ise_potential_per_llvm_bb first!"

            def lookup_supported(df, func_name, bb_name):
                filtered = df[(df["func_name"] == func_name) & (df["bb_name"] == bb_name)]
                # TODO: add missing!
                if len(filtered) == 0:
                    # TODO: check?
                    return 0.0
                assert len(filtered) == 1
                rel_supported_count = filtered["supported_rel_count"].iloc[0]
                return rel_supported_count

            rel_supported_count = lookup_supported(ise_potential_per_llvm_bb_df, func_name, bb_name)
            rel_supported_weight = rel_supported_count * rel_weight
            if rel_supported_weight < min_supported_weight:
                continue
        if min_instrs is not None:
            if num_instrs < min_instrs:
                continue
        func_name = row["func_name"]
        bb_name = row["bb_name"]
        freq = row["freq"]
        if file2funcs_df is not None:
            files = lookup_files(file2funcs_df, func_name)
            if len(files) == 0:
                logger.info(
                    "Falling back to symbol_map based file extraction for '%s'",
                    func_name,
                )
                assert symbol_map_df is not None  # TODO: skip if missing
                file = file_from_symbol_map(func_name, symbol_map_df)
                if file is None:
                    logger.info("Fallback failed!")
                else:
                    logger.info("Fallback succeeded!")
                # func not conatined in file2funcs?
            else:
                assert len(files) == 1
                file = files[0]
        else:
            file = None
        choice = {
            "func_name": func_name,
            "file": file,
            "bb_name": bb_name,
            "rel_weight": rel_weight,
            "num_instrs": num_instrs,
            "freq": freq,
        }
        choices.append(choice)
        sum_weights += rel_weight
        if sum_weights >= threshold:
            break
        if max_num is not None:
            if len(choices) >= max_num:
                break
    # print("sum_weights", sum_weights)
    # print("choices", choices, len(choices))
    choices_df = pd.DataFrame(choices)
    # print("choices_df", choices_df)

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "list",
        "by": __name__,
    }

    artifact = TableArtifact("choices", choices_df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    choose_bbs(
        sess,
        threshold=args.threshold,
        min_weight=args.min_weight,
        min_supported_weight=args.min_supported_weight,
        max_num=args.max_num,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--min-weight", type=float, default=0.01)
    parser.add_argument("--min-supported-weight", type=float, default=None)
    parser.add_argument("--max-num", type=int, default=None)
    parser.add_argument("--threshold", type=float, default=0.9)
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import shutil
from typing import Union, Optional, Dict, Any
from enum import IntFlag, auto
from pathlib import Path

import pickle
import pandas as pd
import networkx as nx

from isaac_toolkit.logging import get_logger
from .config import ArtifactsSettings


logger = get_logger()


class ArtifactFlag(IntFlag):
    # INPUT = auto()
    # OUTPUT = auto()
    # TEMP = auto()
    TABLE = auto()
    GRAPH = auto()
    ELF = auto()
    INSTR_TRACE = auto()
    SOURCE = auto()
    M2ISAR = auto()
    PYTHON = auto()
    DISASS = auto()
    TRACE = auto()


def filter_artifacts(artifacts, func):
    return list(filter(func, artifacts))


# class ExportedArtifact()


class Artifact:

    def __init__(
        self,
        name: str,
        path: Optional[Path],
        flags: Optional[ArtifactFlag] = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        self.name = name
        self.path = Path(path) if path is not None else path
        self._flags = flags if flags is not None else ArtifactFlag(0)
        self.attrs = attrs if attrs is not None else {}
        self.autoload = autoload
        self.changed: bool = True
        self.imported: bool = False
        if autoload:
            self.load()

    @classmethod
    def from_dict(cls, data: dict):
        name = data.get("name", None)
        assert name is not None
        dest = data.get("dest", None)
        assert dest is not None
        flags = data.get("flags", None)
        assert flags is not None
        attrs = data.get("attrs", None)
        assert attrs is not None
        flags_ = ArtifactFlag(flags)
        ret = cls(name, path=dest, flags=flags_, attrs=attrs)
        ret.changed = False
        return ret

    def to_dict(self):
        return {
            "name": self.name,
            "flags": int(self.flags),
            "type": type(self).__name__,
            "dest": str(self.path),
            "hash": self.hash,
            "attrs": self.attrs,
        }

    @property
    def exported(self):
        return self.path is not None and self.path.is_file()

    @property
    def flags(self):
        return self._flags

    @property
    def hash(self):
        return "N/A"  # TODO

    def __repr__(self):
        return f"{type(self).__name__}({self.name}, attrs={self.attrs})"

    def summary(self):
        return f"{self.name}: {self}"

    def update(self):
        self.changed = True

    # @property
    # def is_input(self):
    #     return self.flags & ArtifactFlag.INPUT

    # @property
    # def is_output(self):
    #     return self.flags & ArtifactFlag.OUTPUT

    def _save(self, dest: Path, artifacts_settings: ArtifactsSettings = None):
        raise NotImplementedError("Artifact._save() impl missing")

    def save(self, dest: Path, artifacts_settings: ArtifactsSettings = None):
        if not self.changed and self.exported:
            logger.debug("Artifact '%s' is already exported and unchanged", self.name)
            return
        logger.debug("Exporting artifact '%s' to '%s'", self.name, dest)
        self._save(dest, artifacts_settings=artifacts_settings)
        if self.path is None or not self.path.is_file():
            self.path = dest
        self.changed = False

    def _load(self, dest: Path):
        raise NotImplementedError("Artifact._load() impl missing")

    def load(self, source: Optional[Path] = None):
        if self.imported:
            logger.debug("Artifact '%s' is already imported", self.name)
            # TODO: check sha?
            return
        if source is None:
            assert self.path is not None
            source = self.path
        logger.debug("Loading artifact '%s' from '%s'", self.name, source)
        self._load(source)
        self.imported = True
        self.changed = False


class FileArtifact(Artifact):

    def __init__(
        self,
        name: str,
        path: Optional[Union[str, Path]] = None,
        flags: ArtifactFlag = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        super().__init__(name, path=path, flags=flags, attrs=attrs, autoload=autoload)
        # TODO: content/raw?

    def _save(self, dest: Path, artifacts_settings: ArtifactsSettings = None):
        del artifacts_settings
        if isinstance(dest, str):
            dest = Path(dest)
        if dest.resolve() == self.path.resolve():
            return
        shutil.copyfile(self.path, dest)


class ElfArtifact(FileArtifact):

    @property
    def flags(self):
        return super().flags | ArtifactFlag.ELF


class DisassArtifact(FileArtifact):

    @property
    def flags(self):
        return super().flags | ArtifactFlag.DISASS


class SourceArtifact(FileArtifact):

    @property
    def flags(self):
        return super().flags | ArtifactFlag.SOURCE


class PythonArtifact(Artifact):

    def __init__(
        self,
        name: str,
        data=None,
        path: Optional[Path] = None,
        flags: ArtifactFlag = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        super().__init__(name, path=path, flags=flags, attrs=attrs, autoload=autoload)
        self._data = data

    @property
    def data(self):
        if self._data is None:
            self.load()
        return self._data

    @property
    def flags(self):
        return super().flags | ArtifactFlag.PYTHON

    def _save(self, dest: Path, artifacts_settings: ArtifactsSettings = None):
        del artifacts_settings
        with open(dest, "wb") as f:
            pickle.dump(self.data, f)

    def _load(self, source: Path):
        with open(source, "rb") as f:
            data = pickle.load(f)
        self._data = data


class M2ISARArtifact(PythonArtifact):

    def __init__(
        self,
        name: str,
        model=None,
        path: Optional[Path] = None,
        flags: ArtifactFlag = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        super().__init__(name, data=model, flags=flags, attrs=attrs, autoload=autoload)

    @property
    def flags(self):
        return super().flags | ArtifactFlag.M2ISAR

    @property
    def model(self):
        return self.data


class TableArtifact(PythonArtifact):

    def __init__(
        self,
        name: str,
        df: Optional[pd.DataFrame] = None,
        path: Optional[Path] = None,
        flags: ArtifactFlag = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        super().__init__(name, data=df, path=path, flags=flags, attrs=attrs, autoload=autoload)

    @property
    def flags(self):
        return super().flags | ArtifactFlag.TABLE

    @property
    def df(self):
        return self.data

    def _save(self, dest: Path, artifacts_settings: ArtifactsSettings = None):
        # TODO: csv support?
        if self.flags & ArtifactFlag.INSTR_TRACE:
            fmt = artifacts_settings.instr_trace.fmt
            engine = artifacts_settings.instr_trace.engine
            compression_method = artifacts_settings.instr_trace.compression_method
            compression_level = artifacts_settings.instr_trace.compression_level
        elif self.flags & ArtifactFlag.TRACE:
            fmt = artifacts_settings.trace.fmt
            engine = artifacts_settings.trace.engine
            compression_method = artifacts_settings.trace.compression_method
            compression_level = artifacts_settings.trace.compression_level
        else:
            fmt = artifacts_settings.table.fmt
            engine = artifacts_settings.table.engine
            compression_method = artifacts_settings.table.compression_method
            compression_level = artifacts_settings.table.compression_level
        if fmt == "pickle":
            compression = None
            if compression_method is not None:
                if compression_method == "gzip":
                    compression = {"method": compression_method}
                    if compression_level is not None:
                        compression = {**compression, "compresslevel": compression_level}
                else:
                    assert compression_level is None, "Compression level only supported for pickle+gzip"
                    compression = compression_method

            self.df.to_pickle(dest, compression=compression)
        elif fmt == "parquet":
            assert compression_level is None, "Compression level only supported for pickle+gzip"
            self.df.to_parquet(dest, engine=engine, compression=compression_method)
        else:
            raise ValueError(f"Unsupported format: {self.fmt}")

    def _load(self, source: Path):
        if ".pkl" in source.name:
            df = pd.read_pickle(source, compression="infer")
        elif ".parquet" in source.name:
            # df = pd.read_parquet(source, engine=self.engine)
            df = pd.read_parquet(source)
        else:
            raise ValueError(f"Unsupported format: {source.suffix}")
        self._data = df

    def summary(self):
        return f"{self.name}: pd.DataFrame(shape={self.df.shape}, columns={list(self.df.columns)})"


class TraceArtifact(TableArtifact):

    @property
    def flags(self):
        return super().flags | ArtifactFlag.TRACE


class InstrTraceArtifact(TraceArtifact):

    @property
    def flags(self):
        return super().flags | ArtifactFlag.INSTR_TRACE


class GraphArtifact(PythonArtifact):

    def __init__(
        self,
        name: str,
        graph: Optional[nx.Graph] = None,
        path: Optional[Path] = None,
        flags: ArtifactFlag = None,
        attrs: Optional[Dict[str, Any]] = None,
        autoload: bool = False,
    ):
        super().__init__(name, data=graph, path=path, flags=flags, attrs=attrs, autoload=autoload)

    @property
    def flags(self):
        return super().flags | ArtifactFlag.GRAPH

    @property
    def graph(self):
        return self.data

    def summary(self):
        return f"{self.name}: {self.data}"



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def get_effective_footprint_df(trace_df, func2pc_df, footprint_df):
    df = footprint_df.copy()
    trace_df_unique = trace_df[["pc"]].drop_duplicates()
    df["Used"] = False
    func2pc_df[["start", "end"]] = func2pc_df["pc_range"].apply(pd.Series)
    # print("footprint_df", footprint_df)
    for index, row in footprint_df.iterrows():
        # print("index", index)
        # print("row", row)
        func_name = row["func"]
        matches = func2pc_df.where(lambda x: x["func"] == func_name).dropna()
        # print("matches", matches)
        assert len(matches) > 0
        for pc_range in matches["pc_range"].values:
            start_pc, end_pc = pc_range
            if end_pc < 0:
                continue
            matches = trace_df_unique.where(lambda x: x["pc"] >= start_pc).dropna()
            matches = matches.where(lambda x: x["pc"] < end_pc).dropna()
            if len(matches) > 0:
                df.loc[index, "Used"] = True
                break
    # bytes_before = df["bytes"].sum()
    df = df[df["Used"]]
    bytes_after = df["bytes"].sum()
    df["eff_rel_bytes"] = df["bytes"] / bytes_after
    # rel = bytes_after / bytes_before
    # print("bytes", bytes_before, bytes_after, rel)
    # print("df", df)
    # input("*")
    return df


def track_unused_functions(sess: Session, force: bool = False):
    logger.info("Tracking unused functions...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    mem_footprint_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_footprint"
    )
    assert len(mem_footprint_artifacts) == 1
    mem_footprint_artifact = mem_footprint_artifacts[0]

    effective_footprint_df = get_effective_footprint_df(
        trace_artifact.df, func2pc_artifact.df, mem_footprint_artifact.df
    )
    # print("effective_footprint_df", effective_footprint_df)
    # input("@")

    attrs = {
        "trace": trace_artifact.name,
        # "elf": ?
        "kind": "footprint",
        "by": __name__,
    }

    effective_mem_footprint_artifact = TableArtifact("effective_mem_footprint", effective_footprint_df, attrs=attrs)
    sess.add_artifact(effective_mem_footprint_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    track_unused_functions(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2024 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import shutil
import subprocess

# import yaml
import logging
import argparse
from typing import Optional, Union, List
from pathlib import Path

from isaac_toolkit.session import Session


DEFAULT_DOCKER_IMAGE = "isax-tools-integration-env:latest"
SUPPORTED_CORES = []


logger = logging.getLogger("nailgun")


def run_nailgun_hls(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    index_files: Optional[List[Union[str, Path]]] = None,
    tools_dir: Union[str, Path] = None,
    mount_dir: Union[str, Path] = None,
    docker_image: Optional[str] = None,
    fix_permissions: bool = True,  # Only for docker mode
    # gurubi_license_file: TODO
    isaxes: Optional[List[str]] = None,
    label: Optional[str] = None,
    core_name: Optional[str] = None,
    copy_files: bool = True,
    ilp_solver: str = "gurobi",
    cell_library: Optional[str] = None,
    # cell_library = "$(pwd)/cfg/longnail/library.yaml",
    clock_ns: int = 100,
    schedule_timeout: int = 10,
    refine_timeout: int = 10,
    sched_algo_ms: bool = False,
    sched_algo_pa: bool = False,
    sched_algo_ra: bool = False,
    sched_algo_mi: bool = False,
    share_resources: bool = False,
    mlir_file: Optional[Union[str, Path]] = None,
    force: bool = False,
):
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    if not isinstance(tools_dir, Path):
        tools_dir = Path(tools_dir)
    assert tools_dir.is_dir()
    assert core_name is not None
    assert core_name in SUPPORTED_CORES
    use_docker = docker_image is not None
    subdir = "docker" if use_docker else "local"
    base_dir = workdir / subdir
    hls_dir = base_dir / "hls"
    # syn_dir = base_dir / "syn"
    hls_dir.mkdir(exist_ok=True, parents=True)
    if isaxes is None:
        isaxes = []
    if label is None:
        label = "baseline" if len(isaxes) == 0 else ("shared" if share_resources else "default")
    output_dir = hls_dir / label / "output"
    if output_dir.is_dir():
        assert force, f"Directory already exists: {output_dir}. Use --force or different --label."
        logger.info("Cleaning up old output dir: %s (--force)", output_dir)
        shutil.rmtree(output_dir)
    else:
        output_dir.mkdir(parents=True)
    nailgun_env = {
        "OUTPUT_PATH": output_dir,
        "CONFIG_PATH": output_dir / ".config",
        "SIM_EN": "n",
        "SKIP_AWESOME_LLVM": "y",
        "CORE": core_name,
    }
    use_ol2 = False
    if use_ol2:
        assert mount_dir is not None
        assert Path(mount_dir).is_dir()
        ol2_config_template = mount_dir / "/cfg/openlane/minimal_config_fast.json"
        ol2_until_step = "OpenROAD.Floorplan"
        ol2_target_freq = 20
        ol2_target_util = 20
        nailgun_env.update(
            {
                "OL2_ENABLE": "y",
                "OL2_CONFIG_TEMPLATE": ol2_config_template,
                "OL2_UNTIL_STEP": ol2_until_step,
                "OL2_TARGET_FREQ": ol2_target_freq,
                "OL2_TARGET_UTIL": ol2_target_util,
            }
        )

    def bool_helper(x):
        assert isinstance(x, bool)
        return "y" if x else "n"

    if share_resources:
        assert mlir_file is not None
        assert Path(mlir_file).is_file()
        # TODO: sed -e "s/lil.enc_immediates/lil.sharing_group = 1, lil.enc_immediates/g" $WORK/docker/hls/default/output/mlir/ISAX_ISAAC_EN.mlir > $WORK/docker/hls/ISAX_ISAAC_EN_shared.mlir
        nailgun_env["MLIR_ENTRY_POINT_PATH"] = mlir_file

    elif len(isaxes) == 0:
        nailgun_env["NO_ISAX"] = "y"
    else:
        nailgun_env.update(
            {
                "ISAXES": ",".join(map(lambda x: x.upper(), isaxes)),
            }
        )
    if len(isaxes) > 0:
        nailgun_env.update(
            {
                "LN_ILP_SOLVER": ilp_solver.upper(),
                "CLOCK_TIME": clock_ns,
                "SCHEDULE_TIMEOUT": schedule_timeout,
                "REFINE_TIMEOUT": refine_timeout,
                "SCHED_ALGO_MS": bool_helper(sched_algo_ms),
                "SCHED_ALGO_PA": bool_helper(sched_algo_pa),
                "SCHED_ALGO_RA": bool_helper(sched_algo_ra),
                "SCHED_ALGO_MI": bool_helper(sched_algo_mi),
            }
        )
        if cell_library is not None:
            nailgun_env["CELL_LIBRARY"] = cell_library
    nailgun_command = "make gen_config ci"
    if use_docker:
        command = f"docker run -it --rm -v {tools_dir}:/isax_tools"
        if mount_dir is not None:
            command += f" -v {mount_dir}:{mount_dir}"
        command += f" {docker_image}"
        prepare = "cd /isax-tools/nailgun"
        finalize = f"chmod -R 777 {hls_dir}"
        nailgun_command_env = " ".join([f"{key}={val}" for key, val in nailgun_env.items()])
        command += f' "{prepare} && {nailgun_command_env} {nailgun_command} && {finalize}"'
        print("command", command)
        subprocess.run(command, shell=True)
    else:
        env = os.environ.copy()
        env.update(nailgun_env)
        command = nailgun_command
        cwd = tools_dir
        print("command", command)
        subprocess.run(command, cwd=cwd, env=env, shell=True)
    if copy_files:
        raise NotImplementedError


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    run_nailgun_hls(
        sess,
        force=args.force,
        workdir=args.workdir,
        tools_dir=args.tools_dir,
        docker_image=args.docker,
        isaxes=args.isaxes.split(";") if args.isaxes is not None else None,
        index_files=args.index.split(";") if args.index is not None else None,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--docker", type=str, default=None, const=DEFAULT_DOCKER_IMAGE, nargs="?")
    parser.add_argument("--tools-dir", type=str, required=True)  # TODO: define via settings?
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--isaxes", type=str, default=None)
    # parser.add_argument("--index", type=str, default=None)
    # gurubi_license_file: TODO
    # isaxes: Optional[List[str]] = None,
    # label: Optional[str] = None,
    # core_name: Optional[str] = None,
    # copy_files: bool = True,
    # ilp_solver: str = "gurobi",
    # cell_library: Optional[str] = None,
    # # cell_library = "$(pwd)/cfg/longnail/library.yaml",
    # clock_ns: int = 100,
    # schedule_timeout: int = 10,
    # refine_timeout: int = 10,
    # sched_algo_ms: bool = False,
    # sched_algo_pa: bool = False,
    # sched_algo_ra: bool = False,
    # sched_algo_mi: bool = False,
    # share_resources: bool = False,
    # mlir_file: Optional[Union[str, Path]] = None,

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# from pandarallel import pandarallel

# pandarallel.initialize()


def process_df(df):
    # print("process_df", len(df), df.memory_usage(deep=True).sum())
    # return None

    assert len(df.columns) == 4, "Excpected 4 columns"
    df = df.rename(columns={0: "pc", 1: "?", 2: "is_branch", 3: "size"})
    # df["is_branch"] = df["is_branch"].astype("category")
    # print("A", time.time())

    # df["pc"] = df["pc"].parallel_apply(lambda x: int(x, 0))
    df["pc"] = df["pc"].apply(lambda x: int(x, 0))
    # df["pc"] = df["pc"].apply(lambda x: int(x, 0))
    # df["pc"] = df["pc"].apply(lambda x: int(x, 0))
    # df["pc"] = df["pc"].map(lambda x: int(x, 0))
    # df["pc"] = df["pc"].astype(int, base=0)
    # df["pc"] = pd.to_numeric(df["pc"])
    df["size"] = df["size"].astype(int)
    df["size"] = df["size"].astype("category")
    df.drop(columns=["?"], inplace=True)
    # df.drop(columns=["is_branch"], inplace=True)

    # df["bytecode"] = df[["pc", "size"]].apply(lambda x: lookup_bytecode(x["pc"], x["size"]), axis=1)
    # df["instr"] = df[["pc", "bytecode", "size"]].apply(
    #     lambda x: lookup_name(x["bytecode"], pc=x["pc"], size=x["size"]), axis=1
    # )
    # df["instr"] = df["instr"].astype("category")
    # print("A")
    # unique_pc_size = df[["pc", "size"]].drop_duplicates()
    # unique_pc_size["bytecode"] = unique_pc_size.apply(
    #     lambda x: fetcher.read_word_at_pc(x["pc"], size=x["size"]), axis=1
    # )
    # print("B")

    # def disassemble_row(row):
    #     return disassemble_word(md, int(row["pc"]), int(row["bytecode"]), size=int(row["size"]), operands=False)

    # unique_pc_size["instr"] = unique_pc_size.apply(disassemble_row, axis=1)
    # print("unique_pc_size", unique_pc_size)
    # input(">>")
    # print("C")
    # df = df.merge(unique_pc_size, on=["pc", "size"], how="left")
    # df["bytecode"] = pd.to_numeric(df["bytecode"], downcast="unsigned")
    # df["instr"] = df["instr"].astype("category")
    # print("D")

    # print("pc2bytecode", len(pc2bytecode))
    # print("df", df.head(), df.columns, df.dtypes, df.memory_usage())
    # print("ret")
    return df



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# from pandarallel import pandarallel

# pandarallel.initialize()

import io
import itertools
from pathlib import Path
from typing import Optional
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from tqdm import tqdm
import pandas as pd


DEFAULT_CHUNK_SIZE = 2**22


def parse_and_process(chunk_bytes, process_func, sep: str = ",", names="infer", operands: bool = False, header=None):
    df = pd.read_csv(io.BytesIO(chunk_bytes), header=header, sep=sep, names=names)
    return process_func(df, operands=operands)


def chunk_iter(path, chunk_size=2**22):
    f = open(path, "rb")

    with f:
        buf = b""
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            buf += data
            last_nl = buf.rfind(b"\n")
            if last_nl == -1:
                continue
            yield buf[: last_nl + 1]
            buf = buf[last_nl + 1 :]
        if buf:
            yield buf


def parse_instr_trace(
    input_file: Path,
    process_func: callable,
    num_workers: Optional[int] = None,
    progress: bool = False,
    executor: str = "process_pool",
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    sep: str = ",",
    names="infer",
    header=None,
    operands: bool = False,
):
    dfs = []
    executor_map = {
        "thread_pool": ThreadPoolExecutor,
        "process_pool": ProcessPoolExecutor,
    }
    executor_cls = executor_map.get(executor)
    assert executor_cls is not None, f"Unsupported Executor: {executor}"
    with executor_cls(max_workers=num_workers) as executor:  # tune workers
        dfs_ = list(
            tqdm(
                executor.map(
                    parse_and_process,
                    chunk_iter(input_file, chunk_size=chunk_size),
                    itertools.repeat(process_func),
                    itertools.repeat(sep),
                    itertools.repeat(names),
                    itertools.repeat(operands),
                    itertools.repeat(header),
                ),
                disable=not progress,
            )
        )
        dfs += dfs_
    df = pd.concat(dfs, axis=0)
    if "instr" in df.columns:
        df["instr"] = df["instr"].astype("category")
    if "size" in df.columns:
        df["size"] = df["size"].astype("category")
    if "pc" in df.columns:
        df["pc"] = pd.to_numeric(df["pc"], downcast="unsigned")
    if "bytecode" in df.columns:
        df["bytecode"] = pd.to_numeric(df["bytecode"], downcast="unsigned")
    df.reset_index(drop=True, inplace=True)
    return df



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import sys
import argparse
from pathlib import Path
from typing import Optional

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import InstrTraceArtifact, TraceArtifact
from isaac_toolkit.logging import get_logger, set_log_level
from .utils import parse_instr_trace, DEFAULT_CHUNK_SIZE

logger = get_logger()


def process_etiss_trace_df(df, operands: bool = False):
    df["pc"] = df["pc"].apply(lambda x: int(x, 0))
    df["pc"] = pd.to_numeric(df["pc"])
    df[["instr", "rest"]] = df["rest"].str.split(" # ", n=1, expand=True)
    df["instr"] = df["instr"].apply(lambda x: x.strip())
    df["instr"] = df["instr"].astype("category")
    df[["bytecode", "operands"]] = df["rest"].str.split(" ", n=1, expand=True)

    def detect_size(bytecode):
        if bytecode[:2] == "0x":
            return len(bytecode[2:]) // 2
        elif bytecode[:2] == "0b":
            return len(bytecode[2:]) // 8
        else:
            assert len(set(bytecode)) == 2
            return len(bytecode) // 8

    df["size"] = df["bytecode"].apply(detect_size)
    df["size"] = df["size"].astype("category")
    df["bytecode"] = df["bytecode"].apply(
        lambda x: (int(x, 16) if "0x" in x else (int(x, 2) if "0b" in x else int(x, 2)))
    )
    df["bytecode"] = pd.to_numeric(df["bytecode"])

    def convert(x):
        ret = {}
        for y in x:
            if len(y.strip()) == 0:
                continue
            assert "=" in y
            k, v = y.split("=", 1)
            assert k not in ret
            ret[k] = int(v)
        return ret

    if operands:
        df["operands"] = df["operands"].apply(lambda x: convert(x[1:-1].split(" | ")))
    else:
        df.drop(columns=["operands"], inplace=True)
    df.drop(columns=["rest"], inplace=True)
    return df


def load_instr_trace(
    sess: Session,
    input_file: Path,
    force: bool = False,
    operands: bool = False,
    progress: bool = False,
    num_workers: Optional[int] = None,
    executor: str = "process_pool",
    chunk_size: int = DEFAULT_CHUNK_SIZE,
):
    logger.info("Loading ETISS intruction trace...")
    assert input_file.is_file(), f"File not found: {input_file}"
    name = input_file.name

    df = parse_instr_trace(
        input_file,
        process_etiss_trace_df,
        num_workers=num_workers,
        progress=progress,
        chunk_size=chunk_size,
        executor=executor,
        sep=":",
        names=["pc", "rest"],
        operands=operands,
        header=None,
    )

    attrs = {
        "simulator": "etiss",
        "cpu_arch": "unknown",
        "by": "isaac_toolkit.frontend.instr_trace.etiss",
    }
    if operands:
        operands_trace_df = df[["instr", "operands"]]
        df.drop(columns=["operands"], inplace=True)
        operands_artifact = TraceArtifact("operands_trace", operands_trace_df, attrs=attrs)
        sess.add_artifact(operands_artifact, override=force)
    artifact = InstrTraceArtifact(name, df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    input_file = Path(args.file)
    load_instr_trace(
        sess,
        input_file,
        force=args.force,
        operands=args.operands,
        progress=args.progress,
        executor=args.executor,
        num_workers=args.parallel,
        chunk_size=args.chunk_size,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--operands", action="store_true")
    parser.add_argument("--progress", action="store_true")
    parser.add_argument("--executor", type=str, choices=["process_pool", "thread_pool"], default="process_pool")
    parser.add_argument("--parallel", type=int, default=None)
    parser.add_argument("--chunk-size", type=int, default=DEFAULT_CHUNK_SIZE)
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path
from typing import Optional

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import (
    ArtifactFlag,
    filter_artifacts,
    InstrTraceArtifact,
)
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def trunc_trace(
    sess: Session,
    start_pc: Optional[int] = None,
    end_pc: Optional[str] = None,
    start_func: Optional[str] = None,
    end_func: Optional[str] = None,
    force: bool = False,
):
    logger.info("Truncating trace...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    # print("elf_artifacts", elf_artifacts)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    trace_df = trace_artifact.df
    assert force

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "func2pc")
    if len(func2pc_artifacts) > 0:
        assert len(func2pc_artifacts) == 1
        func2pc_artifact = func2pc_artifacts[0]
        func2pc_df = func2pc_artifact.df
    else:
        func2pc_df = None

    # if start_pc is not None:
    #     assert start_func is None
    # if end_pc is not None:
    #     assert end_func is None

    # TODO: allow start at 0 and/or end at -1

    def lookup_func_pc(func2pc_df: pd.DataFrame, func_name: str):
        assert func2pc_df is not None
        match_df = func2pc_df[func2pc_df["func"] == func_name]
        # print("match_df", match_df)
        assert len(match_df) > 0
        assert len(match_df) == 1
        pc_range = match_df["pc_range"].values[0]
        # print("pc_range", pc_range)
        start_pc = pc_range[0]
        assert start_pc > 0
        return start_pc

    if start_pc is None:
        if start_func is not None:
            start_pc = lookup_func_pc(func2pc_df, start_func)
    if end_pc is None:
        if end_func is not None:
            end_pc = lookup_func_pc(func2pc_df, end_func)

    # print("start_pc", start_pc)
    # print("end_pc", end_pc)
    # TODO: handle multiple calls to start/end func
    def do_trunc(df, start, end):
        if start is not None:
            start_rows = df[df["pc"] == start]
        else:
            start_rows = df.iloc[0:1]
        # print("start_rows", start_rows)
        start_pos = start_rows.index[0]
        # print("start_pos", start_pos)
        # print("temp1", temp1)
        # print("temp1.", temp1.iloc[0])
        # print("temp1.i", temp1.iloc[0].index)
        # print("temp1.i2", temp1.index[0])
        if end is not None:
            end_rows = df[df["pc"] == end]
        else:
            end_rows = df.iloc[-1:-1]
        # print("end_rows", end_rows)
        if len(end_rows) == 0:
            end_pos = df.index[-1]
        else:
            end_pos = end_rows.index[0]
        # print("end_pos", end_pos)
        # print("temp2", temp2)
        # print("temp2.", temp2.iloc[0])
        # print("temp2.i", temp2.iloc[0].index)
        # print("temp2.i2", temp2.index[0])
        # print("AAA", df.iloc[start_pos:end_pos])
        # print("BBB", df.loc[start_pos:end_pos])
        # return df.iloc[start_pos:end_pos]
        return df.loc[start_pos:end_pos]

    trunc_df = do_trunc(trace_df, start_pc, end_pc)
    assert len(trunc_df) > 0
    # print("trunc_df", len(trunc_df))

    # attrs = {
    #     "trace": trace_artifact.name,
    #     "kind": "mapping",
    #     "by": __name__,
    # }
    attrs = trace_artifact.attrs.copy()
    attrs["truncated"] = True
    artifact = InstrTraceArtifact(trace_artifact.name, trunc_df, attrs=attrs)
    # input("555")

    # pc2bb_artifact = TableArtifact(f"pc2bb", pc2bb, attrs=attrs)
    # sess.add_artifact(pc2bb_artifact, override=force)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    trunc_trace(
        sess,
        start_pc=args.start_pc,
        end_pc=args.end_pc,
        start_func=args.start_func,
        end_func=args.end_func,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--start-pc", type=int, default=None)
    parser.add_argument("--end-pc", type=int, default=None)
    parser.add_argument("--start-func", type=str, default=None)
    parser.add_argument("--end-func", type=str, default=None)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from typing import Optional
from pathlib import Path

from isaac_toolkit.session import Session

# from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.frontend.elf.riscv import load_elf
from isaac_toolkit.frontend.linker_map import load_linker_map
from isaac_toolkit.frontend.instr_trace.tgc import (
    load_instr_trace as load_tgc_instr_trace,
)
from isaac_toolkit.frontend.instr_trace.etiss import (
    load_instr_trace as load_etiss_instr_trace,
)
from isaac_toolkit.frontend.instr_trace.etiss_new import (
    load_instr_trace as load_etiss_perf_instr_trace,
)
from isaac_toolkit.frontend.instr_trace.spike import (
    load_instr_trace as load_spike_instr_trace,
)
from isaac_toolkit.frontend.disass.objdump import load_disass

# from isaac_toolkit.frontend.compile_commands.json import load_compile_commands_json

# logger = get_logger()

logger = logging.getLogger()


def load_artifacts(
    sess: Session,
    elf_file: Optional[Path] = None,
    linker_map_file: Optional[Path] = None,
    instr_trace_file: Optional[Path] = None,
    disass_file: Optional[Path] = None,
    force: bool = False,
    progress: bool = False,
):
    logger.info("Loading RVF Demo artifacts...")
    if elf_file:
        load_elf(sess, elf_file, force=force)
    if linker_map_file:
        load_linker_map(sess, linker_map_file, force=force)
    if instr_trace_file:
        assert "_instrs." in instr_trace_file.name
        simulator = instr_trace_file.name.split("_instrs.", 1)[0]
        instr_trace_frontends = {
            "tgc": load_tgc_instr_trace,
            "etiss": load_etiss_instr_trace,
            "etiss_perf": load_etiss_perf_instr_trace,
            "spike": load_spike_instr_trace,
            "spike_rv32": load_spike_instr_trace,
            "spike_rv64": load_spike_instr_trace,
            "spike_bm": load_spike_instr_trace,
        }
        load_instr_trace = instr_trace_frontends.get(simulator)
        assert load_instr_trace is not None
        operands = False  # TODO: store operands in extra artifact!
        load_instr_trace(
            # sess, instr_trace_file, force=force, progress=progress, operands=operands
            sess,
            instr_trace_file,
            force=force,
            operands=operands,
        )
    if disass_file:
        load_disass(sess, disass_file, force=force)
    # load_compile_commands_json(sess, compile_commands_file, force=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)

    # set_log_level(console_level=args.log, file_level=args.log)
    def path_helper(x):
        if x is None:
            return x
        return Path(x)

    load_artifacts(
        sess,
        elf_file=path_helper(args.elf),
        linker_map_file=path_helper(args.linker_map),
        instr_trace_file=path_helper(args.instr_trace),
        disass_file=path_helper(args.disass),
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--elf", default=None)
    parser.add_argument("--linker-map", default=None)
    parser.add_argument("--instr-trace", default=None)
    parser.add_argument("--disass", default=None)
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--progress", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse

import pandas as pd


def get_seal5_score_df(
    seal5_status_df: pd.DataFrame,
    seal5_status_compact_df: pd.DataFrame,
):
    seal5_score_data = []
    seal5_pattern_gen_status_df = seal5_status_df[
        seal5_status_df["pass"] == "generate_passes.pattern_gen.behav_to_pat"
    ][["instr", "status"]]
    seal5_passes_status_df = seal5_status_compact_df[["instr", "status"]]
    # print("seal5_passes_status_df", seal5_passes_status_df)
    for instr_name in seal5_status_df["instr"].unique():
        # print("instr_name", instr_name)
        passes_status = seal5_passes_status_df[seal5_passes_status_df["instr"] == instr_name]
        assert len(passes_status) == 1
        passes_status = passes_status["status"].iloc[0]
        passes_score = 1.0 if passes_status == "good" else (0.5 if passes_status == "ok" else 0.0)

        pattern_gen_status = seal5_pattern_gen_status_df[seal5_pattern_gen_status_df["instr"] == instr_name]
        assert len(pattern_gen_status) == 1
        pattern_gen_status = pattern_gen_status["status"].iloc[0]
        # print("pattern_gen_status", pattern_gen_status)
        pattern_gen_score = 1.0 if pattern_gen_status == "success" else -1.0
        # print("pattern_gen_score", pattern_gen_score)

        new = {"instr": instr_name, "pattern_gen_score": pattern_gen_score, "passes_score": passes_score}
        seal5_score_data.append(new)

    def calc_seal5_score(x):
        return x.min()

    seal5_score_df = pd.DataFrame(seal5_score_data)
    seal5_score_df["seal5_score"] = seal5_score_df[["pattern_gen_score", "passes_score"]].apply(
        calc_seal5_score, axis=1
    )
    return seal5_score_df


def calc_seal5_score(seal5_status_csv, seal5_status_compact_csv, output=None):
    seal5_status_df = pd.read_csv(seal5_status_csv)
    seal5_status_compact_df = pd.read_csv(seal5_status_compact_csv)

    seal5_score_df = get_seal5_score_df(seal5_status_df, seal5_status_compact_df)

    if output is None:
        print(seal5_score_df)
    else:
        seal5_score_df.to_csv(output, index=False)


def handle(args):
    calc_seal5_score(args.seal5_status_csv, args.seal5_status_compact_csv, output=args.output)


def get_parser():
    parser = argparse.ArgumentParser(description="Combine scores into single CSV")
    parser.add_argument("--seal5-status-csv", required=True, help="Seal5 status csv file")
    parser.add_argument("--seal5-status-compact-csv", required=True, help="Seal5 status csv file")
    parser.add_argument("-o", "--output", default=None, help="Output CSV file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.frontend.elf.riscv import load_elf
from isaac_toolkit.frontend.linker_map import load_linker_map
from isaac_toolkit.frontend.instr_trace.etiss import (
    load_instr_trace as load_etiss_instr_trace,
)
from isaac_toolkit.frontend.instr_trace.spike import (
    load_instr_trace as load_spike_instr_trace,
)
from isaac_toolkit.frontend.disass.objdump import load_disass
from isaac_toolkit.frontend.compile_commands.json import load_compile_commands_json

logger = get_logger()


def load_mlonmcu_exported_run(sess: Session, run_dir: Path, force: bool = False, progress: bool = False):
    logger.info("Loading MLonMCU exported run...")
    assert run_dir.is_dir(), f"MLonMCU run dir does not exist: {run_dir}"
    elf_file = run_dir / "generic_mlonmcu"
    load_elf(sess, elf_file, force=force)
    linker_map_file = run_dir / "mlif" / "generic" / "linker.map"  # TODO: move to real artifacts?
    load_linker_map(sess, linker_map_file, force=force)
    # TODO: load report via frontend?
    # TODO: load initializer?
    report_file = run_dir / "report.csv"
    assert report_file.is_file()
    report_df = pd.read_csv(report_file)
    assert len(report_df) == 1
    assert "Target" in report_df.columns
    target = report_df["Target"].iloc[0]
    instr_trace_frontends = {
        "etiss": load_etiss_instr_trace,
        "spike": load_spike_instr_trace,
        "spike_rv32": load_spike_instr_trace,
        "spike_rv64": load_spike_instr_trace,
    }
    load_instr_trace = instr_trace_frontends.get(target)
    assert load_instr_trace is not None, f"Frontend lookup failed for target: {target}"
    instr_trace_file = run_dir / f"{target}_instrs.log"
    operands = True  # TODO: store operands in extra artifact!
    load_instr_trace(sess, instr_trace_file, force=force, progress=progress, operands=operands)
    dump_file = run_dir / "generic_mlonmcu.dump"
    load_disass(sess, dump_file, force=force)
    compile_commands_file = run_dir / "mlif" / "compile_commands.json"
    load_compile_commands_json(sess, compile_commands_file, force=force)


def load_artifacts(sess: Session, run_dir: Path, force: bool = False, progress: bool = False):
    logger.info("Loading ISAAC Demo artifacts...")
    load_mlonmcu_exported_run(sess, run_dir, force=force, progress=progress)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    run_dir = Path(args.exported_run)
    load_artifacts(sess, run_dir, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("exported_run")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--progress", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import tarfile
import tempfile
import shutil
from pathlib import Path


def extract_zstd(archive: Path, out_path: Path):
    """Extract a .zst archive (Zstandard-compressed tar)."""
    archive = Path(archive).expanduser()
    out_path = Path(out_path).expanduser().resolve()
    out_path.mkdir(parents=True, exist_ok=True)

    if sys.version_info >= (3, 14):
        from compression.zstd import ZstdFile

        with ZstdFile(archive, "rb") as f_in:
            with out_path.open("wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
    else:
        _legacy_decompress_zstd_stream(archive, out_path)


def _legacy_decompress_zstd_stream(archive: Path, out_path: Path) -> None:
    from zstandard import ZstdDecompressor

    with tempfile.TemporaryFile(suffix=".tar") as ofh:
        with archive.open("rb") as ifh:
            ZstdDecompressor().copy_stream(ifh, ofh)
        ofh.seek(0)
        with tarfile.open(fileobj=ofh) as z:
            z.extractall(out_path)



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

import networkx as nx
from networkx.drawing.nx_agraph import write_dot
from isaac_toolkit.logging import get_logger

logger = get_logger()


def graph_to_file(graph, dest, fmt="auto"):
    if not isinstance(dest, Path):
        dest = Path(dest)
    if fmt == "auto":
        fmt = dest.suffix[1:].upper()
    prog = "dot"
    # TODO: support pkl
    if fmt == "DOT":
        write_dot(graph, dest)
    elif fmt in ["PDF", "PNG"]:
        graph = nx.nx_agraph.to_agraph(graph)
        graph.draw(str(dest), prog=prog)
        graph.close()
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


def memgraph_to_nx(results):
    graph = nx.MultiDiGraph()
    nodes = list(results.graph()._nodes.values())
    # print("nodes", nodes)
    for node in nodes:
        # print("node", node)
        if len(node._labels) > 0:
            label = list(node._labels)[0]
        else:
            label = "?!"
        name = node._properties.get("name", "?")
        graph.add_node(
            int(node.element_id), key=int(node.element_id), xlabel=label, label=name, properties=node._properties
        )

    rels = list(results.graph()._relationships.values())
    for rel in rels:
        label = rel.type
        # graph.add_edge(
        #     rel.start_node.element_id, rel.end_node.element_id, key=rel.element_id,
        #     label=label, type=rel.type, properties=rel._properties,
        # )
        graph.add_edge(
            int(rel.start_node.element_id),
            int(rel.end_node.element_id),
            key=int(rel.element_id),
            label=label,
            type=rel.type,
            properties=rel._properties,
        )
    return graph


# def calc_inputs(G, sub, ignore_const: bool = False):
def calc_inputs(G, sub):
    # print("calc_inputs", G, sub)
    # print("G.nodes", G.nodes)
    inputs = []
    constants = []
    sub_nodes = sub.nodes
    # print("sub_nodes", sub_nodes)
    for node in sub_nodes:
        # print("node", node, G.nodes[node].get("label"))
        ins = G.in_edges(node)
        # print("ins", ins)
        for in_ in ins:
            # print("in_", in_, G.nodes[in_[0]].get("label"))
            src = in_[0]
            # print("src", src, G.nodes[src].get("label"))
            # print("src in sub_nodes", src in sub_nodes)
            # print("src not in inputs", src not in inputs)
            # op_type = G.nodes[src]["properties"]["op_type"]
            if not (src in sub_nodes) and (src not in inputs):
                # print("IN")
                if G.nodes[src]["properties"]["op_type"] == "constant":
                    constants.append(src)
                else:
                    inputs.append(src)
    # print("ret", ret)
    return len(inputs), inputs, len(constants), constants


def calc_outputs(G, sub):
    # print("calc_outputs", sub)
    ret = 0
    sub_nodes = sub.nodes
    # print("sub_nodes", sub_nodes)
    outputs = []
    for node in sub_nodes:
        # print("node", node, G.nodes[node].get("label"))
        if G.nodes[node]["properties"]["op_type"] == "output":
            # print("A")
            # print("OUT2")
            ret += 1
            if node not in outputs:
                outputs.append(node)
        else:
            # print("B")
            outs = G.out_edges(node)
            # print("outs", outs)
            for out_ in outs:
                # print("out_", out_, G.nodes[out_[0]].get("label"))
                dst = out_[1]
                # print("dst", dst, G.nodes[dst].get("label"))
                if dst not in sub_nodes:
                    # print("OUT")
                    ret += 1
                    if node not in outputs:
                        outputs.append(node)
    # print("ret", ret)
    return ret, outputs


def get_instructions(sub):
    ret = []
    sub_nodes = sub.nodes
    for node in sub_nodes:
        # print("node", node)
        node_data = sub.nodes[node]
        # print("node_data", node_data)
        node_properties = node_data["properties"]
        name = node_properties["name"]
        op_type = node_properties["op_type"]
        if op_type == "input":
            continue
        ret.append(name)
    return ret


def calc_weights(sub):
    weights = []
    freqs = []
    sub_nodes = sub.nodes
    for node in sub_nodes:
        # print("node", node)
        node_data = sub.nodes[node]
        # print("node_data", node_data)
        node_properties = node_data["properties"]
        op_type = node_properties["op_type"]
        if op_type == "input":
            continue
        instr_freq = node_properties.get("instr_freq", None)
        if instr_freq is None:
            return None, None
        instr_rel_weight = node_properties.get("instr_rel_weight", None)
        if instr_rel_weight is None:
            return None, None
        freqs.append(instr_freq)
        weights.append(instr_rel_weight)
    # print("weights", weights)
    # print("freqs", freqs)
    # crossBB = True
    crossBB = False
    if not crossBB:
        assert len(set(weights)) == 1
    total_weight = sum(weights)
    if not crossBB:
        assert len(set(freqs)) == 1
        freq = freqs[0]
    else:
        freq = max(freqs)
    # print("total_weight", total_weight)
    # print("freq", freq)
    # input("?!")
    return total_weight, freq


def calc_weights_iso(graph, nodes):
    weights = []
    freqs = []
    assert len(nodes) == len(set(nodes))
    for node in nodes:
        # print("node", node)
        node_data = graph.nodes[node]
        # print("node_data", node_data)
        node_properties = node_data["properties"]
        op_type = node_properties["op_type"]
        if op_type == "input":
            continue
        instr_freq = node_properties.get("instr_freq", None)
        if instr_freq is None:
            return None, None
        instr_rel_weight = node_properties.get("instr_rel_weight", None)
        if instr_rel_weight is None:
            return None, None
        freqs.append(instr_freq)
        weights.append(instr_rel_weight)
    # print("weights", weights)
    # print("freqs", freqs)
    # crossBB = True
    crossBB = False
    if not crossBB:
        # print("weights", weights)
        assert len(set(weights)) == 1
    total_weight = sum(weights)
    if not crossBB:
        assert len(set(freqs)) == 1
        freq = freqs[0]
    else:
        freq = max(freqs)
    # print("total_weight", total_weight)
    # print("freq", freq)
    # input("?!")
    return total_weight, freq



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def map_llvm_bbs_new(sess: Session, force: bool = False, sort_by: str = "weight"):
    logger.info("Mapping LLVM BBs...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    # print("trace_artifact", trace_artifact)
    trace_df = trace_artifact.df
    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs"
    )  # TODO: optional or different pass
    assert len(llvm_bbs_artifacts) == 1
    llvm_bbs_artifact = llvm_bbs_artifacts[0]
    llvm_bbs_df = llvm_bbs_artifact.df.copy()
    llvm_bbs_df[["start", "end"]] = llvm_bbs_df["pcs"].apply(pd.Series)
    # print("llvm_bbs_df", llvm_bbs_df)
    pc_counts = trace_df["pc"].value_counts()  # TODO: get from historgram?
    total_weight = 0
    for index, row in llvm_bbs_df.sort_values("start").iterrows():
        # print("row", row)
        start = row["start"]
        end = row["end"]
        size = row["size"]
        num_instrs = row["num_instrs"]
        if size == 0:
            continue
        # assert num_instrs != 0  # TODO: handle pcs missing in capstone disassembly?
        assert size > 0, "Encountered basic block with negative size"
        # num_instrs = row["num_instrs"]

        def get_bb_freq_weight(df, start, end, num_instrs):
            # print("get_bb_freq", start, end, num_instrs)
            if num_instrs == 0:
                return 0, 0.0
            assert num_instrs > 0
            matches = df.where(lambda x: x["pc"] >= start).dropna()
            # print("m1", matches)
            matches = matches.where(lambda x: x["pc"] < end).dropna()
            # print("m2", matches)
            # print("matches", matches)
            count = len(matches)
            # print("count", count)
            bb_count = count // num_instrs
            # print("bb_count", bb_count)
            bb_weight = bb_count * num_instrs
            # print("bb_weight", bb_weight)
            return bb_count, bb_weight

        def get_bb_freq_weight2(pc_counts, start, end, num_instrs):
            if num_instrs == 0:
                return 0, 0.0
            assert num_instrs > 0
            matches = pc_counts[pc_counts.index >= start].dropna()
            matches = matches.where(matches.index < end).dropna()
            count = sum(matches)
            bb_count = count // num_instrs
            bb_weight = bb_count * num_instrs
            return bb_count, bb_weight

        # freq, weight = get_bb_freq_weight(trace_df, start, end, num_instrs)
        freq, weight = get_bb_freq_weight2(pc_counts, start, end, num_instrs)
        llvm_bbs_df.loc[index, "freq"] = freq
        llvm_bbs_df.loc[index, "weight"] = weight
        llvm_bbs_df.loc[index, "num_instrs"] = num_instrs
        total_weight += weight
    trace_length = len(trace_df)
    # coverage = total_weight / trace_length
    # print("trace_length", trace_length)
    # print("total_weight", total_weight)
    # print("coverage", coverage)
    llvm_bbs_df.sort_values(sort_by, inplace=True, ascending=False)
    llvm_bbs_df["rel_weight"] = llvm_bbs_df["weight"] / trace_length
    # input(">")

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    artifact = TableArtifact("llvm_bbs_new", llvm_bbs_df, attrs=attrs)
    # print("artifact", artifact)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    map_llvm_bbs_new(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import humanize
import pandas as pd


from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact, FileArtifact, PythonArtifact

from .report_utils import (
    save_pdf_report,
    JUPYTER_CSS,
)
from .sess_disk_usage import size_df_to_markdown, size_df_to_html, get_file_size


def get_artifact_size(artifact):
    if isinstance(artifact, FileArtifact):
        path = artifact.path
        mem = get_file_size(path)
        return mem
    if isinstance(artifact, TableArtifact):
        df = artifact.df
        mem = df.memory_usage(deep=True)
        total_mem = mem.sum() if not isinstance(mem, int) else mem
        return total_mem
    if isinstance(artifact, PythonArtifact):
        data = artifact.data
        mem = sys.getsizeof(data)
        return mem
    raise NotImplementedError(f"Unhandled type: {type(artifact).__name__}")


def format_size(size: float, raw: bool = False):
    if raw:
        return str(size)  # No suffix?
    natural_size = humanize.naturalsize(size)
    return natural_size


def generate_sess_mem_usage(
    sess, output=None, fmt="md", detailed=False, portable=False, style=False, topk=10, force=False, raw=False
):
    # Determine output dir
    out_dir = Path(output) if output else (sess.directory / "reports")
    out_dir.mkdir(exist_ok=True)

    artifacts = sess.artifacts
    artifacts_data = []
    for artifact in artifacts:
        artifact_name = artifact.name
        artifact_kind = str(type(artifact).__name__)
        artifact_size = get_artifact_size(artifact)
        artifacts_data.append((artifact_name, artifact_kind, artifact_size))
    artifacts_df = pd.DataFrame(artifacts_data, columns=["Artifact", "Kind", "Size"])
    total_artifacts_size = artifacts_df["Size"].sum()
    total_artifacts_size_str = format_size(total_artifacts_size, raw=raw)
    artifacts_df = artifacts_df.sort_values("Size", ascending=False)
    if topk is not None:
        artifacts_df = artifacts_df.iloc[:topk]

    # print("Sorted Artifacts DF", artifacts_df)

    if fmt in ("md", "txt"):
        content = "# Session Memory Usage Report\n\n"
        content = "## Overview\n\n"
        content += f"Total Artifacts Mem: {total_artifacts_size_str}\n\n"
        content += "## Artifacts\n\n"
        if topk is not None:
            content += f"TOPK={topk}\n\n"
        content += size_df_to_markdown(artifacts_df, cols=["Size"], raw=raw)
        content += "\n"
        body = content
        ext = "md" if fmt == "md" else "txt"
        outfile = out_dir / f"sess_mem_usage_report.{ext}"
        outfile.write_text(body, encoding="utf-8")
    elif fmt in ("html", "pdf"):
        parts = []
        parts.append("<html><head>")
        if style:
            parts.append(JUPYTER_CSS)
        parts.append("</head><body>")
        parts.append("<h1>Session Memory Usage Report</h1>")
        parts.append("<h2>Overview</h2>")
        parts.append(f"<p>Total Artifacts Size: {total_artifacts_size_str}</p>")
        parts.append("<h2>Artifacts</h2>")
        if topk is not None:
            parts.append(f"TOPK={topk}")
        table_content = size_df_to_html(artifacts_df, cols=["Size"], raw=raw)
        table_content = table_content.replace("<table id=", "<table border='1' class='dataframe dataframe' id=")
        parts.append(table_content)
        parts.append("</body></html>")
        body = "\n".join(parts)
        ext = "html" if fmt == "html" else "pdf"
        outfile = out_dir / f"sess_mem_usage_report.{ext}"
        if fmt == "html":
            outfile.write_text(body, encoding="utf-8")
        else:
            save_pdf_report(body, outfile)
    else:
        assert False, f"Unsupported fmt: {fmt}"

    print(f"[isaac_toolkit.report] Runtime report written to {outfile}")  # TODO: logging!


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    generate_sess_mem_usage(
        sess,
        output=args.out,
        fmt=args.fmt,
        detailed=args.detailed,
        portable=args.portable,
        style=args.style,
        topk=args.topk,
        raw=args.raw,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser(description="Generate runtime report from ISAAC session.")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--out", help="Custom output directory (default: SESSION/reports)")
    parser.add_argument("--fmt", choices=["md", "txt", "html", "pdf"], default="md", help="Report format")
    parser.add_argument("--detailed", action="store_true", help="Include detailed breakdowns")
    parser.add_argument("--portable", action="store_true", help="Embed plots as base64 (only for HTML output)")
    parser.add_argument("--style", action="store_true", help="Use custom CSS for HTML output")
    parser.add_argument("--raw", action="store_true", help="Output raw sizes instead of humanized ones")
    parser.add_argument("--topk", type=int, default=10, help="Limit number of table rows")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities to generate Reports."""
import base64
from pathlib import Path

from weasyprint import HTML


def save_pdf_report(html_str, out_path):
    """Convert HTML string into PDF file."""
    HTML(string=html_str).write_pdf(out_path)


JUPYTER_CSS = """
<style>
table.dataframe {
  border-collapse: collapse;
  border: 1px solid #aaa;
  font-family: sans-serif;
  margin-bottom: 1em;
}
table.dataframe th, table.dataframe td {
  border: 1px solid #aaa;
  padding: 4px 8px;
  text-align: right;
}
table.dataframe th {
  background-color: #f2f2f2;
}
table.dataframe tr:nth-child(even) {
  background-color: #fafafa;
}
table.dataframe tr:hover {
  background-color: #f5f5f5;
}

body {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue",
               Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji";
  margin: 20px;
  color: #333;
  line-height: 1.5;
}

h1, h2 {
  font-weight: 600;
  font-family: inherit;
  border-bottom: 2px solid #eaeaea;
  padding-bottom: 4px;
  margin-top: 1.5em;
  margin-bottom: 0.75em;
}

h1 {
  font-size: 1.8em;
  color: #222;
}

h2 {
  font-size: 1.4em;
  color: #444;
}

table.dataframe {
  margin-top: 1em;
}
</style>
"""


def encode_image_base64(path: Path) -> str:
    """Convert image into bas64 for HTML embedding."""
    with open(path, "rb") as f:
        data = f.read()
    return base64.b64encode(data).decode("utf-8")


def format_hex(x):
    """Convert addresses into hexadecimal."""
    return hex(int(x))


def monospace_addresses_html(val):
    """Wrap hex addresses into HTML code tags."""
    if isinstance(val, str) and val.startswith("0x"):
        return f"<code>{val}</code>"
    return val


def monospace_addresses_md(val):
    """Wrap hex addresses into Markdown code tags."""
    if isinstance(val, str) and val.startswith("0x"):
        return f"`{val}`"
    return val



import sys
import argparse
import pandas as pd
import yaml


def annotate_util_score(index, util_score_csv, output=None, inplace=False, in_prefix="", out_prefix=""):
    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    # print("candidates", candidates)

    scores_df = pd.read_csv(util_score_csv)
    # print("scores_df", scores_df)
    # assert len(candidates) == len(scores_df)

    for i, candidate in enumerate(candidates):
        # print("i", i)
        name = candidate["properties"]["InstrName"]
        instr_row = scores_df[scores_df["instr"] == name]
        # print("instr_row", instr_row)
        assert len(instr_row) == 1
        util_score = float(instr_row[f"{in_prefix}util_score"].iloc[0])
        metrics = candidate.get("metrics", {})
        # print("metrics", metrics)
        metrics[f"{out_prefix}util_score"] = util_score
        # print("metrics2", metrics)
        candidate["metrics"] = metrics

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    combined_index_data["candidates"] = candidates

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


def handle(args):
    annotate_util_score(
        args.index,
        args.util_score_csv,
        output=args.output,
        inplace=args.inplace,
        in_prefix=args.in_prefix,
        out_prefix=args.out_prefix,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--util-score-csv", required=True, help="TODO")
    parser.add_argument("--in-prefix", default="", help="TODO")
    parser.add_argument("--out-prefix", default="", help="TODO")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.arch.riscv import riscv_branch_instrs, riscv_return_instrs
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


# def find_sub_bbs(bbs, first_pc, last_pc):
#     ret = set()
#     for bb_ in bbs:
#         if bb_.first_pc >= first_pc and bb_.last_pc <= last_pc:
#             ret.add(bb_)
#     return ret
#
# def find_super_bbs(bbs, first_pc, last_pc):
#     ret = set()
#     for bb_ in bbs:
#         if bb_.first_pc <= first_pc and bb_.last_pc >= last_pc:
#             ret.add(bb_)
#     return ret
#
# def get_diff_range(bb, first_pc, last_pc):
#     if bb.last_pc == last_pc:
#         first_pc = first_pc
#         last_pc = bb.first_pc - 4  # TODO: this breaks for sz=2
#     else:
#         raise NotImplementedError
#     return (first_pc, last_pc)


class BasicBlock(object):
    _instances = {}
    # _children = defaultdict(list)
    # _parents = defauktdict(set)

    @staticmethod
    def get_instances():
        return list(BasicBlock._instances.values())

    def __new__(
        cls,
        first_pc: int,
        last_pc: int,
        num_instrs: int,
        size: int,
        end_instr: str,
        func: str,
    ):
        # print("__new__", first_pc, last_pc, end_instr, func)
        # print("_instances", BasicBlock._instances)
        key = (first_pc, last_pc, num_instrs, size, end_instr, func)
        # print("key", key)
        instance = cls._instances.get(key)
        # children = cls._children.get(key)
        # print("instance", instance)
        # print("children", children)
        if instance is None:
            # if children is not None:
            #     assert len(children) > 1
            #     for x in children:
            #         x._freq += 1
            #     return children[0]
            # sub_bbs = find_sub_bbs(cls.get_instances(), first_pc, last_pc)
            # new_children = []
            # if len(sub_bbs) > 0:
            #     print("sub_bbs", sub_bbs)
            #     assert len(sub_bbs) == 1
            #     sub_bb = list(sub_bbs)[0]
            #     new_children.append(sub_bb)
            #     sub_bb._freq += 1
            #     diff_range = get_diff_range(sub_bb, first_pc, last_pc)
            #     print("diff_range", diff_range)
            #     # new_bb = BasicBlock(first_pc=diff_range[0], last_pc=diff_range[1], end_instr="?", func=None)
            #     # new_bb._freq += sub_bb._freq
            #     first_pc = diff_range[0]
            #     last_pc = diff_range[1]
            #     end_instr = "?"
            #     func = None
            #     # print("new_bb", new_bb)
            #     input("SUB")
            #     pass
            # super_bbs = find_sub_bbs(cls.get_instances(), first_pc, last_pc)
            # if len(super_bbs) > 0:
            #     print("super_bbs", super_bbs)
            #     assert len(super_bbs) == 1
            #     input("SUPER")
            #     pass
            instance = super().__new__(cls)
            # new_children.append(instance)
            # if len(new_children) > 1:
            #     for x in new_children:
            #         cls._children[key].append(x)
            #     for x in new_children[:-1]:
            #         cls._parents[key].append(instance)
            cls._instances[key] = instance
            instance.__initialized = False
            instance._freq = 0
        instance._freq += 1
        return instance

    def __init__(
        self,
        first_pc: int,
        last_pc: int,
        num_instrs: int,
        size: int,
        end_instr: str,
        func: str,
    ) -> None:
        if not self.__initialized:
            self.first_pc = first_pc
            self.last_pc = last_pc
            self.num_instrs = num_instrs
            self.size = size
            self.func = func
            self.end_instr = end_instr
            self.__initialized = True

    def __repr__(self) -> str:
        return (
            f"start:{hex(self.first_pc)}, end:{hex(self.last_pc)}, num_instrs:{self.num_instrs}, "
            f"size:{self.size}, end_instr:{self.end_instr}, func:{self.func}\n"
        )

    def __eq__(self, other) -> bool:
        # print("__eq__", self, other, end="")
        if not isinstance(other, BasicBlock):
            # print(" -> False1")
            return False
        ret = self.first_pc == other.first_pc and self.last_pc == other.last_pc and self.func == other.func
        # print(f" -> {ret}2")
        return ret

    def __hash__(self) -> int:
        return hash((self.first_pc, self.last_pc, self.end_instr, self.func))

    def get_freq(self) -> int:
        return self._freq


def collect_bbs(trace_df):
    logger.info("Collecting BBs...")
    # print("trace_df", len(trace_df))
    # input("{}{}{}{}")
    first_pc = None
    # TODO: make this generic!
    # bbs = []
    prev_pc = None
    prev_size = None
    # prev_instr = None
    bb_instrs = []
    bb_size = 0
    for row in trace_df.itertuples(index=False):
        # print("row", row)
        pc = row.pc
        instr = row.instr
        instr = instr.strip()  # TODO: fix in frontend
        # def detect_compressed(name):
        #     print("name", name)
        #     if name[0] == "c":
        #         return True
        #     return False
        # is_compressed = detect_compressed(instr)
        is_compressed = row.size == 2
        sz = 2 if is_compressed else 4
        # print("sz", sz)

        if prev_pc:
            # print("pc", pc)
            # print("prev_pc", prev_pc)
            # print("prev_size", prev_size)
            step = pc - prev_pc
            # print("step", step)
            if step in [2, 4] and step == prev_size:
                pass
            else:
                # is_jump = True
                if first_pc is None:
                    pass
                else:
                    # assert False, f"Sub basic block not found at: pc = {prev_pc:x} -> {pc:x}"
                    logger.warning("Detected potential trap @ pc = 0x%x -> 0x%x", prev_pc, pc)
                    # input("OOPS")
                    if True:
                        func = None
                        bb = BasicBlock(
                            first_pc=first_pc,
                            last_pc=prev_pc,
                            num_instrs=len(bb_instrs),
                            size=bb_size,
                            end_instr=instr,
                            func=func,
                        )
                        bb_instrs = []
                        bb_size = 0
                        first_pc = pc
                        # if bb.get_freq() == 1:
                        #     bbs.append(bb)

        # At the first pc of a basic block
        if first_pc is None:
            first_pc = pc

        if instr in riscv_branch_instrs + riscv_return_instrs:
            # func = self.find_func_name(pc)
            func = None
            bb = BasicBlock(
                first_pc=first_pc,
                last_pc=pc,
                num_instrs=len(bb_instrs),
                size=bb_size,
                end_instr=instr,
                func=func,
            )
            bb_instrs = []
            bb_size = 0
            # self.func_set.add(func)
            # if bb.get_freq() == 1:
            #     # sub_bbs = find_overlapping_bbs(bb, bbs)
            #     # print("sub_bbs", sub_bbs)
            #     # if len(sub_bbs) > 0:
            #     #     assert len(sub_bbs) == 1
            #     #     for sub_bb in sub_bbs:
            #     #         sub_bb._freq += bb._freq
            #     #         diff_range = get_diff_range(bb, sub_bb)
            #     #         print("diff_range", diff_range)
            #     #         new_bb = BasicBlock(first_pc=diff_range[0], last_pc=diff_range[1], end_instr="?", func=None)
            #     #         new_bb._freq = bb._freq
            #     #         BasicBlock.instances
            #     #         print("new_bb", new_bb)
            #     #     input("OVERLAP")
            #     bbs.append(bb)
            first_pc = None
        prev_pc = pc
        # prev_instr = instr
        prev_size = sz
        bb_instrs.append(instr)
        bb_size += sz  # incorrect, need to be before bb creation?
    if first_pc is not None:
        func = None
        # bb = BasicBlock(first_pc=first_pc, last_pc=prev_pc, end_instr=instr, func=func)
        # if bb.get_freq() == 1:
        #     bbs.append(bb)
    bbs = BasicBlock.get_instances()
    # print("bbs", bbs)
    bbs_data = []
    for bb in bbs:
        # print("bb", bb, dir(bb))
        start = bb.first_pc
        end = bb.last_pc
        freq = bb.get_freq()
        # last_size = 4  # TODO: do not hardcode
        # size = end - start + 4
        size = bb.size
        num = bb.num_instrs
        bb_data = {"bb": (start, end), "freq": freq, "size": size, "num_instrs": num}
        bbs_data.append(bb_data)

    bbs_df = pd.DataFrame(bbs_data)
    bbs_df.sort_values("freq", inplace=True, ascending=False)
    bbs_df["weight"] = bbs_df["freq"] * bbs_df["num_instrs"]
    bbs_df["rel_weight"] = bbs_df["weight"] / sum(bbs_df["weight"])
    # print("bbs_df", bbs_df)

    return bbs_df


def analyze_basic_blocks(sess: Session, force: bool = False):
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    # print("elf_artifacts", elf_artifacts)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]

    pc2bb = collect_bbs(trace_artifact.df)
    # print("pc2bb", pc2bb)
    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "func2pc")
    if len(func2pc_artifacts) > 0:
        assert len(func2pc_artifacts) == 1
        func2pc_artifact = func2pc_artifacts[0]
        func2pc_df = func2pc_artifact.df.copy()
        func2pc_df[["start", "end"]] = func2pc_df["pc_range"].apply(pd.Series)

        # print("func2pc_df", func2pc_df)
        def helper(x):
            # print("x", x)
            # print("!", func2pc_df["start"] >= x[0] & func2pc_df["end"] <= x[1])
            matches = func2pc_df[func2pc_df["start"] <= x[0]]
            matches = matches[matches["end"] >= x[1]]
            # print("matches", matches)
            if len(matches) == 0:
                return None
            # print("len(matches)", len(matches))
            # assert len(matches) == 1
            # match_ = matches.iloc[0]
            # func = match_["func"]
            # TODO: assert that alias!
            funcs = set(matches["func"].values)
            # return func
            return funcs

        ANNOTATE_FUNC = True
        if ANNOTATE_FUNC:
            pc2bb["func_name"] = pc2bb["bb"].apply(helper)  # .astype("category")
        pc2bb[["start", "end"]] = pc2bb["bb"].apply(pd.Series)
        del pc2bb["bb"]
        # print("pc2bb", pc2bb)

    attrs = {
        "trace": trace_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    pc2bb_artifact = TableArtifact("pc2bb", pc2bb, attrs=attrs)
    sess.add_artifact(pc2bb_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_basic_blocks(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import argparse
import subprocess
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def get_command_for_file(compile_commands_df: pd.DataFrame, file: str):
    # print("compile_commands_df", compile_commands_df)
    # print("files", compile_commands_df["file"].unique())
    matches = compile_commands_df[compile_commands_df["file_resolved"] == file]
    # print("matches", matches)
    if len(matches) == 0:
        matches = compile_commands_df[compile_commands_df["file_name"] == file]
    assert len(matches) > 0, "Could not find compile commands for file: {file}"
    assert len(matches) == 1, "Too many matches during compile commands lookup for file: {file}"
    command = matches["command"].values[0]
    directory = matches["directory"].values[0]
    return command, directory


def generate_memgraph_cdfg_via_compile_commands(
    sess: Session,
    label: str,
    stage: int = 32,
    force: bool = False,
):
    del force
    logger.info("Generating Memgraph CDFG via compile commands...")
    artifacts = sess.artifacts
    choices_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "choices")
    assert len(choices_artifacts) == 1
    choices_artifact = choices_artifacts[0]
    choices_df = choices_artifact.df
    assert len(choices_df) > 0

    compile_commands_artifacts = filter_artifacts(
        artifacts,
        lambda x: x.flags & ArtifactFlag.TABLE and x.attrs.get("kind") == "compile_commands",
    )
    assert len(compile_commands_artifacts) == 1
    compile_commands_artifact = compile_commands_artifacts[0]
    compile_commands_df = compile_commands_artifact.df
    compile_commands_df["file_resolved"] = compile_commands_df["file"].apply(lambda x: Path(x).resolve())
    compile_commands_df["file_name"] = compile_commands_df["file"].apply(lambda x: Path(x).name)

    # symbol_map_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_map"
    # )
    # symbol_map_df = None
    # if len(symbol_map_artifacts) > 0:
    #     assert len(symbol_map_artifacts) == 1
    #     symbol_map_artifact = symbol_map_artifacts[0]
    #     symbol_map_df = symbol_map_artifact.df

    files = choices_df["file"].unique()

    extra_args = (
        f"-mllvm -cdfg-enable=1 -mllvm -cdfg-memgraph-session={label} -mllvm -cdfg-memgraph-purge=0 "
        f"-mllvm -cdfg-stage-mask={stage}"
    )
    config = sess.config
    memgraph_config = config.memgraph
    memgraph_host = os.environ.get("MEMGRAPH_HOST")
    memgraph_port = os.environ.get("MEMGRAPH_PORT")
    if memgraph_config is not None:
        memgraph_host = memgraph_host or memgraph_config.hostname
        memgraph_port = memgraph_port or memgraph_config.port

    if memgraph_host is not None:
        extra_args += f" -mllvm -cdfg-memgraph-host={memgraph_host}"
    if memgraph_port is not None:
        extra_args += f" -mllvm -cdfg-memgraph-port={memgraph_port}"

    for file in files:
        # print("file", file)
        if file is None:

            logger.warning("Unknown file. Skipping CDFG extraction...")
            continue
        orig_command, directory = get_command_for_file(compile_commands_df, file)
        # print("orig_command", orig_command)
        # print("directory", directory)
        new_command = f"{orig_command} {extra_args}"
        # print("new_command", new_command)
        # new_args = new_command.split(" ")
        # print("args", new_command)
        # print("cwd", directory)
        # input(">")
        subprocess.run(new_command, check=True, shell=True, cwd=directory)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    generate_memgraph_cdfg_via_compile_commands(sess, label=args.label, stage=args.stage, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--label", type=str, default=None)
    parser.add_argument("--stage", type=int, default=8)
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Union


import sys
import logging
import argparse
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def export_pctrace(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
    compress: bool = False,
):
    del force
    artifacts = sess.artifacts

    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    trace_df = trace_artifact.df
    print("trace_df", trace_df.head())
    if "cost" not in trace_df.columns:
        trace_df["cost"] = 1
    COLS = ["pc", "size", "is_branch", "cost"]
    pctrace_df = trace_df[COLS].copy()
    pctrace_df["pc"] = pctrace_df["pc"].map(lambda x: hex(x))
    print("pctrace_df", pctrace_df.head())
    # TODO: handle --force!

    if output is None:
        profile_dir = sess.directory / "outputs"
        profile_dir.mkdir(exist_ok=True)
        out_name = "pctrace.trc"
        output = profile_dir / out_name
    if compress:
        import lz4.frame

        with lz4.frame.open(output, mode="wt") as f:
            pctrace_df.to_csv(f, index=False, header=False)
    else:
        pctrace_df.to_csv(output, index=False, header=False)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    export_pctrace(
        sess,
        output=args.output,
        force=args.force,
        compress=args.compress,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--output", default=None)
    parser.add_argument("--compress", action="store_true", help="Write LZ4 compressed file")
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])
